/*
 * TismTool © Tis Veugen, 2010-2012
 * Simulation, Verification
 *
 * Tsm.h
 * version 3
 */
#ifndef TSM_H
#define TSM_H

#include <iostream>
#include <sstream>
#include <exception>
#include <boost/thread/mutex.hpp>    
using namespace std;

namespace TsmRTE
{
#define TsmAssert( bCondition ) Tsm::Assert( bCondition, #bCondition, __FILE__, __LINE__ )

  class Tsm
  {
  public:
    static bool         mbLogConsole;
    static bool         mbLogFile;
    static int          mLogLevel;
    static bool         mbSeqFile;
    static int          mSeqClassId;
  private:
    static ofstream     mLogWriter;
    static boost::mutex mLogLockMutex;
    static ofstream     mSeqWriter;
    static boost::mutex mSeqLockMutex;

  public:
    static void SetLogConsole( bool bLogConsole )
    {
      mbLogConsole = bLogConsole;
    }

    static void SetLogFile( bool bLogFile )
    {
      mbLogFile = bLogFile;
    }

    static void SetLogLevel( int iLevel )
    {
      mLogLevel = iLevel;
    }

    static void SetSeqFile( bool bSeqFile )
    {
      mbSeqFile = bSeqFile;
    }

    static void Assert( bool bCondition, string msg, string file, int line  )
    {
      if ( ! bCondition )
      {
        stringstream sb;
        sb << "ASSERT @ " << file << "#" << line << " : " << msg;
        cout << sb.str() << endl;
        Log( sb.str() );
      }
    }

    static void Open();
    static void Log( string s );
    static int  SeqClass( string s );
    static void Seq( string s );
    static void Seq2( string s, int n );

    static void Lgl( int iLevel, string s )
    {
      if ( iLevel <= mLogLevel )
      {
        Log( s );
      }
    }

    static void ShowException( string s, exception& e )
    {
      cout << e.what() << endl;
      Lgl( 1, e.what() );
      Lgl( 1, s );
    }

    static void Close();

    static string StackTracestrings( exception& e )
    {
      return "NoTrace";
    }

    static void ReportBox( string s )
    {
      cout << "Report: " << s << endl;
    }

  private:
    static string FreeTismFile( string sName );
  };
}
#endif
