/*
 * TismTool © Tis Veugen, 2010-2012
 * Application, Simulation, Verification
 *
 * TsmCompound.cpp
 * version 1
 */
#include <sstream>
#include "TsmCompound.h"
#include "TsmHandler.h"
#include "TsmRegion.h"
#include "TsmHandlerAction.h"
#include "TsmException.h"
using namespace std;

namespace TsmRTE
{
  void TsmCompound::ExecuteTriggerRegions( TsmHandler& oHandler )
  {
    ExecuteRegions( oHandler, true );
  }

  void TsmCompound::ExecuteRegions( TsmHandler& oHandler, bool bTrigger )
  {
    TsmHandler hAll;

    for ( int i = 0; i < mNrRegions; i++ )
    {
      TsmRegion*  tRegion = mRegions[i];
      TsmHandler h( oHandler.Owner );

      if ( bTrigger )
      {
        h.Trigger = oHandler.Trigger;
        h.Prospect = oHandler.Prospect;
        tRegion->ExecuteTrigger( h );
      }
      else
      {
        h.Action = TsmHandlerAction::ENTER;
        tRegion->ExecuteHandler( h );
      }

      switch ( h.Action )
      {
      case TsmHandlerAction::TRANSITION:
        if ( TsmHandlerAction::hasTransition( hAll.Action ) )
        {
          stringstream sb;
          sb << "9610: Multiple region transitions @ " << this->FullName();
          throw TsmException( sb.str() );
        }
        hAll.Index = h.Index;
        break;
      case TsmHandlerAction::TACTRANSITION:
        if ( TsmHandlerAction::hasTransition( hAll.Action ) )
        {
          stringstream sb;
          sb << "9611: Multiple region transitions @ " << this->FullName();
          throw TsmException( sb.str() );
        }
        hAll.Index = h.Index;
        hAll.TacSubFsm = h.TacSubFsm;
        break;
      default:
        break;
      }
      hAll.Action |= h.Action;
      hAll.bTaken |= h.bTaken;
    }

    HandlerTakeOver( oHandler, hAll );
    oHandler.bTaken |= hAll.bTaken;
  }

  void TsmCompound::HandlerTakeOver( TsmHandler& oHandler, TsmHandler& hAll )
  {
    if ( (hAll.Action & TsmHandlerAction::TACTRANSITION) != 0 )
    {
      oHandler.Action    = TsmHandlerAction::TACTRANSITION;
      oHandler.Index     = hAll.Index;
      oHandler.TacSubFsm = hAll.TacSubFsm;
    }
    else
      if ( (hAll.Action & TsmHandlerAction::TRANSITION) != 0 )
      {
        oHandler.Action = TsmHandlerAction::TRANSITION;
        oHandler.Index  = hAll.Index;
      }
      else
        if ( (hAll.Action & TsmHandlerAction::OK) != 0 )
        {
          oHandler.Action = TsmHandlerAction::OK;
        }
        else
          if ( (hAll.Action & TsmHandlerAction::GUARD) != 0 )
          {
            oHandler.Action = TsmHandlerAction::GUARD;
          }
          else
            if ( hAll.Action == TsmHandlerAction::FINAL )
            {
              oHandler.Action = TsmHandlerAction::FINAL;
            }
            else
              if ( hAll.Action == TsmHandlerAction::NOK )
              {
                oHandler.Action = TsmHandlerAction::NOK;
              }
              else
                if ( TsmHandlerAction::isFinalNok( hAll.Action ) )
                {
                  oHandler.Action = TsmHandlerAction::NOK;
                }
                else
                {
                  oHandler.Action = TsmHandlerAction::NOK;
                  stringstream sb;
                  sb << "TismBUG, HandlerTakeOver action @ " << this->FullName();
                  throw TsmException( sb.str() );
                }
  }

  void TsmCompound::Initial( TsmHandler& oHandler )
  {
    ExecuteRegions( oHandler, false );
  }

  void TsmCompound::ExecuteTrigger( TsmHandler& oHandler )
  {
    int ProspectSelf = GetProspect( oHandler.Trigger->PortId, oHandler.Trigger->Command );
    oHandler.Prospect |= ProspectSelf;

    ExecuteTriggerRegions( oHandler );

    if ( ! oHandler.bTaken )
    {
      // Process the trigger itself, if transition available.
      ExecuteTriggerCompound( oHandler );
    }
  }

  void TsmCompound::ExecuteTriggerCompound( TsmHandler& oHandler )
  {
    // Override for all the compound triggers
    oHandler.Action = TsmHandlerAction::NOK;
  }

  void TsmCompound::ExitRegions( TsmHandler& oHandler )
  {
    for ( int i = 0; i < mNrRegions; i++ )
    {
      mRegions[i]->AtExit( oHandler );
    }
  }

  void TsmCompound::AtExit( TsmHandler& oHandler )
  {
    ExitRegions( oHandler );
    AtMyExit( oHandler );
  }

  string TsmCompound::TsmSnap()
  {
    stringstream sb;

    sb << "[" << this->TsmName << ":";
    for ( int i = 0; i < mNrRegions; i++ )
    {
      sb << "|" << mRegions[i]->TsmSnap();
    }
    sb << "]";
    return sb.str();
  }

  void TsmCompound::TsmCheckVisit()
  {
    for ( int i = 0; i < mNrRegions; i++ )
    {
      mRegions[i]->TsmCheckVisit();
    }
  }
}
