/*
 * TismTool © Tis Veugen, 2010-2012
 * Application, Simulation, Verification
 *
 * TsmRegion.cpp
 * version 2
 */
#include <sstream>
#include "TsmRegion.h"
#include "TsmHandler.h"
#include "TsmHandlerAction.h"
#include "TsmProspect.h"
#include "TsmException.h"
using namespace std;

namespace TsmRTE
{
  void TsmRegion::Initial( TsmHandler& oHandler )
  {
    stringstream sb;
    sb << "9600: Region without Initial state @ " << this->FullName();
    throw TsmException( sb.str() );
  }

  void TsmRegion::ExecuteTrigger( TsmHandler& oHandler )
  {
    if ( mbFinal )
    {
      oHandler.Action = TsmHandlerAction::FINAL;
      return;
    }

    int Prospect = GetProspect( oHandler.Trigger->PortId, oHandler.Trigger->Command );
    switch ( Prospect )
    {
    case TsmProspect::NOK :
      oHandler.Action = TsmHandlerAction::NOK;
      return;
    case TsmProspect::ANY :
      if ( (oHandler.Prospect & TsmProspect::OK) != 0 )
      { /*
         * Execute the trigger at the higher level.
         */
        oHandler.Action = TsmHandlerAction::NOK;
        return;
      }
      break;
    default :
      break;
    }

    mCurrentState->ExecuteTrigger( oHandler );
    mCurrentState->ExecuteHandler( oHandler );
    ExecuteHandler( oHandler );
    mbFinal = ( oHandler.Action == TsmHandlerAction::FINAL );
  }

  void TsmRegion::AtExit( TsmHandler& oHandler )
  {
    if ( NULL != mCurrentState )
    {
      mCurrentState->AtExit( oHandler );
      Reset();
    }
  }

  void TsmRegion::Reset()
  {
    mCurrentState = NULL;
    mbFinal = false;
  }

  string TsmRegion::TsmSnap()
  {
    if ( NULL == mCurrentState )
    {
      return "";
    }
    return mCurrentState->TsmSnap();
  }

  void TsmRegion::TsmCheckVisit()
  {
    for ( int i=0; i<mNrStates; i++ )
    {
      mStates[i]->TsmCheckVisit();
    }
  }
}
