/*
 * TismTool © Tis Veugen, 2010-2012
 * Simulation, Verification
 *
 * TsmCbThread.cpp
 * version 6
 */
#include "TsmCbThread.h"
#include "TsmThreadState.h"
#include "TsmThreadType.h"
#include "TsmSupervisor.h"
#include "TsmEnvelope.h"
#include "TsmRvState.h"
using namespace std;

namespace TsmRTE
{
  void TsmCbThread::Prepare()
  {
    mbHalt = false;
    mThread = new boost::thread( CbWorker, this );
    mSequence = TsmSupervisor::Register( this, mType );
    mExternalQueue->setSeqId( mSequence );
    mInternalQueue->setSeqId( mSequence );
    this->setName( "TismRTE" );
  }

  void TsmCbThread::NativeCall( TsmEnvelope& oEnvelope )
  {
    oEnvelope.getTrigger()->SetDBU();
    mExternalQueue->Put( oEnvelope );
  }

  void TsmCbThread::CbHandle( TsmEnvelope& oEnvelope )
  {
    oEnvelope.getTarget()->ExecuteNative( oEnvelope.getTrigger()->Command );
  }

  void CbWorker( TsmCbThread* oTsmThread )
  {
    TsmEnvelope   lEnvelope;

    oTsmThread->RunningWait();

    while ( ! oTsmThread->mbHalt )
    {
      // Wait for message controlled by Supervisor
      TsmSupervisor::SetState( WAITING_MSG );

      if ( oTsmThread->mExternalQueue->Dequeue( lEnvelope ) )
      {
        bool bDBU = lEnvelope.GetDBU();

        if ( !lEnvelope.getTrigger()->IsForeign )
        {
          Tsm::Seq( "F" );
        }
        try
        {
          oTsmThread->CbHandle( lEnvelope );
        }
        catch( TsmRzVousException )
        { /*
           * Escape from processing.
           */
          TsmSupervisor::SetState( INTERRUPTED );
          Tsm::Lgl( 5, "@q : escape" );
        }
        catch ( exception& e )
        {
          Tsm::ShowException( "@q", e );
        }
        lEnvelope.HandleDBU( bDBU );
      }
      else
      {
        if ( oTsmThread->mInternalQueue->Dequeue( lEnvelope ) )
        {
          bool bDBU = lEnvelope.GetDBU();

          if ( !lEnvelope.getTrigger()->IsForeign )
          {
            Tsm::Seq( "J" );
          }
          try
          {
            oTsmThread->Handle( lEnvelope );
          }
          catch( TsmRzVousException )
          { /*
             * Escape from processing.
             */
            TsmSupervisor::SetState( INTERRUPTED );
            Tsm::Lgl( 5, "@s : escape" );
          }
          catch ( exception& e )
          {
            Tsm::ShowException( "@s", e );
          }
          lEnvelope.HandleDBU( bDBU );
        }
      }
    }
  }  
}
