/*
 * TismTool © Tis Veugen, 2010-2012
 * Simulation, Verification
 *
 * TsmSupervisor.cpp
 * version 4
 */
#include <vector>
#include <boost/date_time.hpp>
#include <boost/thread/thread.hpp>
#include "TsmSupervisor.h"
#include "TsmCbThread.h"
#include "TsmThread.h"
#include "Tsm.h"
#include "TsmSvrHelpers.h"
#include "TsmSvrChecker.h"
using namespace std;

namespace TsmRTE
{
  boost::mutex  TsmSupervisor::LockThreads;
  TsmActivity*  TsmSupervisor::mActivity;
  TsmCbThread*  TsmSupervisor::mCbThread;
  bool          TsmSupervisor::mbVerify = false;

  int       TsmSupervisor::mMaximumQueueSize = 3;

  int       TsmSupervisor::mRandomSeed = 1;
  long      TsmSupervisor::mMaxNrSimLoops = 50;

  void TsmSupervisor::Initialise()
  {
    Tsm::Open();
    mActivity = new TsmActivity();
    Register( NULL, SUPERVISOR );

    mCbThread = new TsmCbThread();
    mCbThread->Prepare();
    mCbThread->Run();
  }

  void TsmSupervisor::Terminate()
  {
    mCbThread->Halt();
    mCbThread->Cease();
    delete mCbThread;
    delete mActivity;
    Tsm::Close();
  }

  TsmIntQueue* TsmSupervisor::GetIntQueue()
  {
    return mCbThread->getIntQueue();
  }

  void TsmSupervisor::Announce()
  {
    mActivity->Announce();
  }

  void TsmSupervisor::Abandon()
  {
    mActivity->Abandon();
  }

  vector<TsmClass*>* TsmSupervisor::GetClasses()
  {
    return TsmSvrChecker::Instance().GetClasses();
  }

  int TsmSupervisor::Register( TsmThread* oTsmThread )
  {
    return Register( oTsmThread, THREAD );
  }

  int TsmSupervisor::Register( TsmThread* oTsmThread, TsmThreadType eType )
  {
    boost::thread::id lThreadBid;
    
    if ( NULL != oTsmThread )
    {
      lThreadBid = oTsmThread->getThread()->get_id();
    }
    else
    {
      lThreadBid = boost::this_thread::get_id();
    }

    TsmSvrChecker& Checker = TsmSvrChecker::Instance();
    return Checker.Register( oTsmThread, eType, lThreadBid );
  }

  void TsmSupervisor::SetState( boost::thread::id ThreadBid )
  {
    SetState( ThreadBid, RUNNING );
  }

  void TsmSupervisor::SetState( TsmThreadState eState )
  {
    SetState( boost::this_thread::get_id(), eState );
  }

  void TsmSupervisor::SetState( boost::thread::id ThreadBid, TsmThreadState eState )
  {
    TsmSvrChecker::Instance().SetState( ThreadBid, eState );
  }

  void TsmSupervisor::NativeEntry( TsmClass* oOwner, const int aNatives[], int NrNatives )
  {
    TsmSvrChecker::Instance().NativeEntry( oOwner, aNatives, NrNatives );
  }

  void TsmSupervisor::NativeExit( TsmClass* oOwner, const int aNatives[], int NrNatives )
  {
    TsmSvrChecker::Instance().NativeExit( oOwner, aNatives, NrNatives );
  }

  string TsmSupervisor::SnappedClasses()
  {
    return TsmSvrChecker::Instance().SnappedClasses();
  }

  string TsmSupervisor::SnappedThreads()
  {
    return TsmSvrChecker::Instance().SnappedThreads();
  }

  string TsmSupervisor::GetCurrentSituation()
  {
    return TsmSvrChecker::Instance().GetCurrentSituation();
  }

  /*
   * Called by Queues.
   */
  bool TsmSupervisor::QueueOverflow( int QueueCount )
  {
    boost::unique_lock<boost::mutex> lock( TsmSupervisor::LockThreads );

    if ( QueueCount >= mMaximumQueueSize )
    {
      TsmSvrChecker::Instance().AddQueueOverflow();
      return true;
    }
    return false;
  }

  /*
   * Called by Fsms.
   */
  bool TsmSupervisor::ChoiceOK( int MaxNrChoices, int ChoiceIndex, string sGuard )
  {
    bool bChoice;

    boost::unique_lock<boost::mutex> lock( TsmSupervisor::LockThreads );
    if ( mbVerify )
    {
      bChoice = TsmSvrChecker::Instance().ChoiceOKVerify( MaxNrChoices, ChoiceIndex, sGuard );
    }
    else
    {
      bChoice = TsmSvrChecker::Instance().ChoiceOKSimulate( MaxNrChoices, ChoiceIndex );
    }
    return bChoice;
  }


  void TsmSupervisor::Verify()
  {
    mbVerify = true;
    TsmSvrChecker::Instance().Verify();
  }

  void TsmSupervisor::ConfigureSimulation( int iSeed, long iNrLoops )
  {
    mRandomSeed = iSeed;
    mMaxNrSimLoops = iNrLoops;
  }

  void TsmSupervisor::Simulate()
  {
    mbVerify = false;
    TsmSvrChecker::Instance().Simulate();
  }

  void TsmSupervisor::Summary()
  {
    TsmSvrChecker::Instance().Summary( mbVerify );
  }

  void TsmSupervisor::Cleanup()
  {
    TsmSvrChecker::Cleanup();
  }
}
