/*
 * TismTool © Tis Veugen, 2010-2012
 * Simulation, Verification
 *
 * TsmSvrChecker.cpp
 * version 11
 */
#include <cstdlib>  // srand
#include <sstream>
#include <set>
#include <boost/shared_ptr.hpp>
#include "TsmSvrChecker.h"
#include "Tsm.h"
#include "TsmException.h"
#include "TsmSupervisor.h"
#include "TsmRecorder.h"
#include "Tsm.h"

using namespace std;

namespace TsmRTE
{
  int            TsmSvrChecker::mThreadId = 0;
  TsmSvrChecker* TsmSvrChecker::SingleInstance;

  int TsmSvrChecker::Register( TsmThread* oTsmThread, TsmThreadType eType, boost::thread::id lThreadBid )
  {
    if ( mThreadDatas.find( lThreadBid ) != mThreadDatas.end() )
    {
      throw TsmException( "TismBUG: Doubly registered thread" );
    }

    TsmThreadData* lThreadData = new TsmThreadData();
    lThreadData->Type       = eType;
    lThreadData->ThreadId   = TsmSvrChecker::mThreadId++;
    lThreadData->pTsmThread = oTsmThread;
    lThreadData->State      = RUNNING;

    switch ( lThreadData->Type )
    {
    case SUPERVISOR:
      MyThreadData = lThreadData;
      break;
    case CB_THREAD:
      CbThreadData = lThreadData;
      break;
    case THREAD:
      mThreadDatas[ lThreadBid ] = lThreadData;
      break;
    }
   
    return lThreadData->ThreadId;
  }

  void TsmSvrChecker::SetState( boost::thread::id ThreadBid, TsmThreadState eState )
  {
    TsmThreadData* lThreadData;

    if ( ThreadBid == CbThreadData->pTsmThread->getThread()->get_id() )
    {
      lThreadData = CbThreadData;
    }
    else
    {
      map<boost::thread::id,TsmThreadData*>::iterator it;
      it = mThreadDatas.find( ThreadBid );
      if ( it == mThreadDatas.end() )
      {
        throw TsmException( "TismBUG: Unregistered thread" );
      }
      lThreadData = it->second;
    }

    lThreadData->State = eState;
    switch ( lThreadData->Type )
    {
    case SUPERVISOR:
      switch ( lThreadData->State )
      {
      case RUNNING:
      case INTERRUPTED:
        throw TsmException( "TismBUG: Supervisor should not report" );
        //break;
      case WAITING_MSG:
      case WAITING_RV:
      case WAITING_RV2:
        throw TsmException( "TismBUG: Supervisor should not wait" );
        //break;
      }
      break;
    case CB_THREAD:
    case THREAD:
      switch ( lThreadData->State )
      {
      case RUNNING:
        break;
      case WAITING_MSG:
        mTriggerSem.Release();
        lThreadData->pTsmThread->TriggerWait();
        break;
      case WAITING_RV:
        mTriggerSem.Release();
        break;
      case WAITING_RV2:
        mTriggerSem.Release();
        lThreadData->pTsmThread->TriggerWait();
        break;
      case INTERRUPTED:
        lThreadData->State = RUNNING;
        break;
      }
      break;
    }
  }

  void TsmSvrChecker::NativesReset()
  {
    map<TsmClass*,TsmNativeLists*>::iterator it;

    for ( it = Natives.begin(); it != Natives.end(); it++ )
    {
      TsmNativeLists* lLists = it->second;
      delete lLists;
    }
    Natives.clear();
  }

  void TsmSvrChecker::NativeEntry( TsmClass* oOwner, const int aNatives[], int NrNatives )
  {
    boost::unique_lock<boost::mutex> lock( LockNatives );
    TsmNativeLists* lLists;
    map<TsmClass*,TsmNativeLists*>::iterator it;

    it = Natives.find( oOwner );

    if ( it == Natives.end() )
    {
      lLists = new TsmNativeLists();
      Natives[ oOwner ] = lLists;
    }
    else
    {
      lLists = it->second;
    }

    for ( int i=0; i<NrNatives; i++ )
    {
      int n = aNatives[i];
      if ( n < 0 )
      {
        lLists->ClientNatives.insert( n );
      }
      else
      {
        lLists->ServerNatives.insert( n );
      }
    }
  }

  void TsmSvrChecker::NativeExit( TsmClass* oOwner, const int aNatives[], int NrNatives )
  {
    boost::unique_lock<boost::mutex> lock( LockNatives );
    TsmNativeLists* lLists;
    map<TsmClass*,TsmNativeLists*>::iterator it;

    it = Natives.find( oOwner );

    if ( it == Natives.end() )
    {
      throw TsmException( "TismBUG: NativeExit " + oOwner->TotalName() );
    }

    lLists = it->second;
    multiset<int>::iterator itn;

    for ( int i=0; i<NrNatives; i++ )
    {
      int n = aNatives[i];
      if ( n < 0 )
      {
        itn = lLists->ClientNatives.find( n );
        if ( itn != lLists->ClientNatives.end() )
        {
          lLists->ClientNatives.erase( itn );
        }
      }
      else
      {
        itn = lLists->ServerNatives.find( n );
        if ( itn != lLists->ServerNatives.end() )
        {
          lLists->ServerNatives.erase( itn );
        }
      }
    }
  }


  void TsmSvrChecker::AddQueueOverflow()
  {
    TsmScenItem* lScenItem = new TsmScenItem();
    lScenItem->Kind = QO;
    CurrentScenario->Add( lScenItem );
    CurrentScenItemIndex++;
    bQueueOverflow = true;
  }

  bool TsmSvrChecker::ChoiceOKVerify( int MaxNrChoices, int ChoiceIndex, string sGuard )
  {
    TsmScenItem* lScenItem;

    if ( CurrentScenItemIndex < (int) CurrentScenario->Items->size() )
    { /*
       * Still replaying.
       */
      boost::shared_ptr<TsmScenItem> SpSi = CurrentScenario->Items->at( CurrentScenItemIndex );
      //TsmAssert( SpSi.get()->Kind == CHOICE );
      if ( SpSi.get()->Index != ChoiceIndex )
      {
        return false;
        /***********/
      }

      CurrentScenItemIndex++;
      if ( CurrentScenItemIndex == (int) CurrentScenario->Items->size() )
      { /*
         * 'Execution' of the last item is happening now.
         */
        TsmRecorder::Enable();
      }
    }
    else
    { /*
       * The scenario has been replayed fully, and
       * now encounters a Choice node.
       */
      //TsmAssert( ChoiceIndex == 1 );
      TsmScenario* lNewScenario;
      for ( int i=2; i<=MaxNrChoices; i++ )
      {
        lScenItem = new TsmScenItem();
        lScenItem->Kind = CHOICE;
        lScenItem->Index = i;
        lNewScenario = CurrentScenario->ExtendChoice( lScenItem );
        FutureScenarios.push( lNewScenario );
      }

      lScenItem = new TsmScenItem();
      lScenItem->Kind = CHOICE;
      lScenItem->Index = 1;
      CurrentScenario->Add( lScenItem );
      CurrentScenItemIndex++;
    }
    Tsm::Seq( "H," + sGuard );
    return true;
  }

  void TsmSvrChecker::Verify()
  {
    vector<TsmScenItem*>  lScenItems;
    bool                  bProxyProtocol = false;
    TsmSnapshot*          lSnapshot;
    TsmSnapdata*          lSnapdata;
    map<string,TsmSnapdata*>::iterator its;

    StartTime = time( NULL );

    if ( Classes.size() == 2 )
    {
      if ( ( Classes[0]->IsTsmProxy() && Classes[1]->IsTsmProxyUser() )
        || ( Classes[1]->IsTsmProxy() && Classes[0]->IsTsmProxyUser() ) )
      {
        bProxyProtocol = true;
      }
    }
    CurrentScenario = new TsmScenario();
    FutureScenarios.push( CurrentScenario );

    bool mbCurrent = false;
    while ( ! FutureScenarios.empty() )
    {
      CurrentScenario = FutureScenarios.top();
      FutureScenarios.pop();
      PastScenarios.push( CurrentScenario );

      if ( mbCurrent )
      {
        mbCurrent = false;

        stringstream ss;
        ss << "S," << CurrentScenario->Id << "," << CurrentScenario->getParentId();
        Tsm::Seq( ss.str() );

        stringstream st;
        st << "Scenario [ " << CurrentScenario->Id << "  :  " << CurrentScenario->getParentId() << " ] -------------------------------------------------------";
        Tsm::Lgl( 2, st.str() );
        Tsm::Lgl( 3, CurrentScenario->toString() );

        if ( CurrentScenario->Items->size() > 0 )
        {
          CheckAllThreadsWaiting();
          boost::shared_ptr<TsmScenItem> SpSi = CurrentScenario->Items->at( CurrentScenItemIndex++ );
          ReplayScenItem( SpSi.get() );
          //TsmAssert( CurrentScenItemIndex == CurrentScenario->Items->size() );
        }
        CheckAllThreadsWaiting();
      }
      else
      {
        Tsm::Seq( "N" );
        Reset();

        stringstream ss;
        ss << "S," << CurrentScenario->Id << "," << CurrentScenario->getParentId();
        Tsm::Seq( ss.str() );

        stringstream st;
        st << "Scenario [ " << CurrentScenario->Id << "  :  " << CurrentScenario->getParentId() << " ] -------------------------------------------------------";
        Tsm::Lgl( 2, st.str() );
        Tsm::Lgl( 3, CurrentScenario->toString() );
        Replay();
      }

      ReplayExtend( CurrentScenario );
      /*
       * Now, no threads have state RV2 !
       */
      lSnapshot = new TsmSnapshot(
        SnappedClasses(),
        CbThreadData->pTsmThread->Snap(),
        SnappedThreads() );
      CurrentScenario->Update( lSnapshot );

      if ( bQueueOverflow )
      {
        bQueueOverflow = false;
        TsmRecorder::QueueOverflow();
        continue;
        /*******/
      }

      Tsm::Seq( "T," + lSnapshot->SnapClasses );
      its = Snapshots.find( lSnapshot->toString() );
      if ( its != Snapshots.end() )
      {
        lSnapdata = its->second;
        Tsm::Lgl( 5, TxtInt( "Equal: ", lSnapdata->Scenario->Id ) );
        Tsm::Seq2( "Q,", lSnapdata->Scenario->Id );
        if ( CurrentScenario->HasCycle() )
        {
          NumberOfCycles++;
          Tsm::Lgl( 4, "Cycle: -------------------------------------------------------" );
          Tsm::Seq( "Y" );
          if ( lSnapdata->NrCalls == TsmRecorder::getNumberOfExtCalls() )
          { /*
             * The number of external calls retrieved from the previous
             * position in lSnapdata is equal to the current number of
             * external calls in the Recorder.
             */
            NumberOfLivelocks++;
            TsmRecorder::Show( "**** LIVELOCK ", lSnapdata->NrRecords );
          }
        }
      }
      else
      {
        Tsm::Lgl( 5, "Rply: -------------------------------------------------------" );
        Tsm::Lgl( 5, "Snap:" + lSnapshot->toString() );
        lSnapdata = new TsmSnapdata( TsmRecorder::getNumberOfRecords(), TsmRecorder::getNumberOfExtCalls(), CurrentScenario );
        Snapshots[ lSnapshot->toString() ] = lSnapdata;

        bool bClient = (CbThreadData->State == WAITING_MSG);
        lScenItems.clear();
        AddScenarioCbThread( lScenItems );
        if ( bProxyProtocol )
        { /*
           * In case of a proxy protocol, native triggers are only
           * allowed in the absense of pending callbacks.
           */
          if ( lScenItems.size() == 0 )
          {
            AddScenarioNatives( lScenItems, bClient );
            AddScenarioThreads( lScenItems );
          }
        }
        else
        {
          AddScenarioNatives( lScenItems, bClient );
          AddScenarioThreads( lScenItems );
        }

        if ( lScenItems.size() == 0 )
        {
          NumberOfDeadlocks++;
          TsmRecorder::Show( "**** DEADLOCK " );
        }
        else
        {
          AddToCurrentScenario( lScenItems );
          mbCurrent = true;
        }
      }
    }

    /*
     * Delete all Snapdata.
     */
    for ( its=Snapshots.begin(); its != Snapshots.end(); its++ )
    {
      delete its->second;
    }

    /*
     * Delete all Snapshots and Scenarios.
     */
    NumberOfScenarios = PastScenarios.size();
    while ( ! PastScenarios.empty() )
    {
      TsmScenario* lTsmScenario = PastScenarios.top();
      PastScenarios.pop();
      delete lTsmScenario->Snapshot;
      delete lTsmScenario;
    }

    /*
     * Cleanup.
     */
    Reset();
    NativesReset();
    Free();
    TsmRecorder::Reset();
    StopTime = time( NULL );
  }


  bool TsmSvrChecker::ChoiceOKSimulate( int MaxNrChoices, int ChoiceIndex )
  {
    return (GetRandomInt( MaxNrChoices + 1 - ChoiceIndex ) == 0);
  }

  void TsmSvrChecker::Simulate()
  {
    vector<TsmScenItem*>  lScenItems;
    TsmScenItem*          lScenItem;
    bool                  bProxyProtocol = false;

    srand( TsmSupervisor::mRandomSeed );
    StartTime = time( NULL );

    if ( Classes.size() == 2 )
    {
      if ( ( Classes[0]->IsTsmProxy() && Classes[1]->IsTsmProxyUser() )
        || ( Classes[1]->IsTsmProxy() && Classes[0]->IsTsmProxyUser() ) )
      {
        bProxyProtocol = true;
      }
    }

    CurrentScenario = new TsmScenario();

    bool bCurrent = false;
    NumberOfLoops = 0;
    do
    {
      if ( ++NumberOfLoops == 0 )
      {
        NumberOfLoops = 1;
      }

      stringstream ss;
      ss << "Loop: " << NumberOfLoops;
      Tsm::Lgl( 3, ss.str() );
      if ( bCurrent )
      {
        bCurrent = false;

        CheckAllThreadsWaiting();
        int r = GetRandomInt( lScenItems.size() );
        lScenItem = lScenItems[r];
        ReplayScenItem( lScenItem );
        CheckAllThreadsWaiting();
      }
      else
      {
        Reset();
        Replay();
      }

      while ( NULL != ( lScenItem = GetThreadDataRV2() ) )
      {
        ReplayScenItem( lScenItem );
        delete lScenItem;
        CheckAllThreadsWaiting();
      }

      bool bClient = ( CbThreadData->State == WAITING_MSG );
      DeleteScenItems( lScenItems );
      AddScenarioCbThread( lScenItems );
      if ( bProxyProtocol )
      { /*
         * In case of a proxy protocol, native triggers are only
         * allowed in the absense of pending callbacks.
         */
        if ( lScenItems.size() == 0 )
        {
          AddScenarioNatives( lScenItems, bClient );
          AddScenarioThreads( lScenItems );
        }
      }
      else
      {
        AddScenarioNatives( lScenItems, bClient );
        AddScenarioThreads( lScenItems );
      }

      if ( lScenItems.size() == 0 )
      {
        NumberOfDeadlocks++;
        TsmRecorder::Show( "**** DEADLOCK " );
      }
      else
      {
        bCurrent = true;
      }
    } while ( NumberOfLoops != TsmSupervisor::mMaxNrSimLoops );

    DeleteScenItems( lScenItems );
    delete CurrentScenario;
    Reset();
    NativesReset();
    Free();
    TsmRecorder::Reset();
    StopTime = time( NULL );
  }

  void TsmSvrChecker::Summary( bool bVerify )
  {
    double DTime = difftime( StopTime, StartTime );
    long duration = (long) (DTime);
    Tsm::Log( "" );
    Tsm::Log( "SUMMARY" );
    Tsm::Log( "-------" );
    if ( bVerify )
    {
      Tsm::Log( TxtInt( "Number of scenarios: ", NumberOfScenarios ) );
      Tsm::Log( TxtInt( "Number of cycles   : ", NumberOfCycles ) );
    }
    else
    {
      Tsm::Log( TxtInt( "Number of loops    : ", NumberOfLoops ) );
    }
    Tsm::Log( TxtInt( "Number of deadlocks: ", NumberOfDeadlocks ) );
    if ( bVerify )
    {
      Tsm::Log( TxtInt( "Number of livelocks: ", NumberOfLivelocks ) );
    }
    Tsm::Log( TxtInt( "Number of unhandled functions  : ", TsmRecorder::NumberOfUnhandledFunctions ) );
    Tsm::Log( TxtInt( "Number of unvisited states     : ", TsmRecorder::NumberOfUnvisitedStates ) );
    Tsm::Log( TxtInt( "Number of unvisited transitions: ", TsmRecorder::NumberOfUnvisitedTransitions ) );
    Tsm::Log( TxtInt( "Number of rendez-vous errors   : ", TsmRecorder::NumberOfSyncReturnErrors ) );
    Tsm::Log( TxtInt( "Number of queue overflows      : ", TsmRecorder::NumberOfQueueOverflows ) );
    Tsm::Log( "Duration: " + GetDuration( duration ) );
    Tsm::Log( "-------" );
    Tsm::Log( "Ext  Int  Thread's maximum queue sizes" );
    Tsm::Log( CbThreadData->pTsmThread->MaxQueueSizes() );

    map<boost::thread::id,TsmThreadData*>::iterator it;
    for ( it=mThreadDatas.begin() ; it != mThreadDatas.end(); it++ )
    {
      TsmThreadData *lThreadData = it->second;

      Tsm::Log( lThreadData->pTsmThread->MaxQueueSizes() );
    }
    Tsm::Log( "-------" );
  }

  string TsmSvrChecker::SnappedClasses()
  {
    stringstream sb;
    int c = 0;
    for( unsigned int i=0; i<Classes.size(); i++ )
    {
      sb << ++c;
      sb << Classes[i]->TsmSnap();
    }
    return sb.str();
  }

  string TsmSvrChecker::SnappedThreads()
  {
    stringstream sb;
    int t = 0;

    map<boost::thread::id,TsmThreadData*>::iterator it;
    for ( it=mThreadDatas.begin() ; it != mThreadDatas.end(); it++ )
    {
      TsmThreadData *lThreadData = it->second;

      sb << " T_" << ++t << ": ";
      sb << lThreadData->pTsmThread->Snap();
    }
    return sb.str();
  }

  string TsmSvrChecker::GetCurrentSituation()
  {
    stringstream sb;

    sb << "Classes:" << endl;
    int c = 0;
    for( unsigned int i=0; i<Classes.size(); i++ )
    {
      sb.width(3);
      sb << ++c << " ";
      sb << Classes[i]->TsmSnap() << endl;
    }

    sb << "Scenario:" << endl;
    TsmScenario* lScenario = PastScenarios.top();
    sb << lScenario->toString();
    return sb.str();
  }

  string TsmSvrChecker::GetThreadStates( string sHeader )
  {
    stringstream sb;

    sb << sHeader << endl;
    sb << MyThreadData->toState() << endl;
    sb << CbThreadData->toState() << endl;

    map<boost::thread::id,TsmThreadData*>::iterator it;
    for ( it=mThreadDatas.begin() ; it != mThreadDatas.end(); it++ )
    {
      TsmThreadData *lThreadData = it->second;

      sb << lThreadData->toState() << endl;
    }
    return sb.str();
  }

  //PRIVATE
  void TsmSvrChecker::DeleteScenItems( vector<TsmScenItem*>& oScenItems )
  {
    for ( unsigned int i=0; i<oScenItems.size(); i++ )
    {
      delete oScenItems[i];
    }
    oScenItems.clear();
  }

  int TsmSvrChecker::GetRandomInt( int iMax )
  {
    return rand() % iMax;
  }

  bool TsmSvrChecker::AllThreadsWaiting()
  {
    if ( RUNNING == CbThreadData->State )
    {
      return false;
    }

    map<boost::thread::id,TsmThreadData*>::iterator it;
    for ( it=mThreadDatas.begin() ; it != mThreadDatas.end(); it++ )
    {
      TsmThreadData *lThreadData = it->second;

      if ( RUNNING == lThreadData->State )
      {
        return false;
      }
    }
    return true;
  }

  bool TsmSvrChecker::AllThreadsSleeping()
  {
    return true;
  }

  void TsmSvrChecker::CheckAllThreadsWaiting()
  {
    while ( ! AllThreadsWaiting() )
    {
      mTriggerSem.WaitOne();
    }
  }

  void TsmSvrChecker::Reset()
  {
    CbThreadData->pTsmThread->TsmReset1();

    map<boost::thread::id,TsmThreadData*>::iterator it;
    for ( it=mThreadDatas.begin() ; it != mThreadDatas.end(); it++ )
    {
      TsmThreadData *lThreadData = it->second;
      lThreadData->pTsmThread->TsmReset1();
    }

    if ( WAITING_RV == CbThreadData->State )
    {
      Tsm::Lgl( 7, "Interrupt Cb" );
      CbThreadData->State = RUNNING;
      CbThreadData->pTsmThread->Interrupt();
    }

    for ( it=mThreadDatas.begin() ; it != mThreadDatas.end(); it++ )
    {
      TsmThreadData *lThreadData = it->second;

      if ( WAITING_RV == lThreadData->State )
      {
        Tsm::Lgl( 7, TxtInt( "Interrupt T_", lThreadData->pTsmThread->mSequence ) );
        lThreadData->State = RUNNING;
        lThreadData->pTsmThread->Interrupt();
      }
    }
    CheckAllThreadsWaiting();

    CbThreadData->pTsmThread->TsmReset2();

    for ( it=mThreadDatas.begin() ; it != mThreadDatas.end(); it++ )
    {
      TsmThreadData *lThreadData = it->second;
      lThreadData->pTsmThread->TsmReset2();
    }

    for( unsigned int i=0; i<Classes.size(); i++ )
    {
      Classes[i]->TsmReset1();
    }
    NativesReset();
  }

  void TsmSvrChecker::Free()
  {
    CbThreadData->pTsmThread->Halt();

    map<boost::thread::id,TsmThreadData*>::iterator it;
    for ( it=mThreadDatas.begin() ; it != mThreadDatas.end(); it++ )
    {
      TsmThreadData *lThreadData = it->second;
      lThreadData->pTsmThread->Halt();
    }
  }

  void TsmSvrChecker::DeleteThreadData()
  {
    delete MyThreadData;
    delete CbThreadData;

    map<boost::thread::id,TsmThreadData*>::iterator it;
    for ( it=mThreadDatas.begin() ; it != mThreadDatas.end(); it++ )
    {
      TsmThreadData *lThreadData = it->second;
      delete lThreadData;
    }
  }

  void TsmSvrChecker::Pause( TsmThreadData* oThreadData )
  {
    oThreadData->pTsmThread->TriggerWait();
  }

  void TsmSvrChecker::Continue( TsmThreadData* oThreadData )
  {
    oThreadData->State = RUNNING;
    oThreadData->pTsmThread->TriggerNotify();

    try
    {
      boost::posix_time::milliseconds ms(1);
      boost::this_thread::sleep( ms );
    }
    catch ( exception )
    {
    }
  }

  void TsmSvrChecker::ReplayNative( TsmScenItem* oItem )
  {
    stringstream st;
    st << "Natv: " << oItem->Index << " @ " << oItem->Receiver->FullName();
    Tsm::Lgl( 6, st.str() );

    if ( oItem->Index < 0 )
    { /*
       * Client native.
       * Executed by CbThread.
       */
      TsmTrigger* lTrigger = new TsmTrigger( -oItem->Index );
      TsmEnvelope lEnvelope( lTrigger, oItem->Receiver );
      CbThreadData->pTsmThread->NativeCall( lEnvelope );
      Continue( CbThreadData );
    }
    else
    { /*
       * Server native.
       * Executed by Supervisor.
       */
      oItem->Receiver->ExecuteNative( oItem->Index );
    }
  }

  /*
   * The state changes :
   *     Item->ThreadData->State = RUNNING;
   * are needed to force the Supervisor to wait for the threads.
   */
  void TsmSvrChecker::ReplayScenItem( TsmScenItem* Item )
  {
    switch ( Item->Kind )
    {
    case RELEASE_CB_MSG :
      //TsmAssert( Item->ThreadData->hasMessage() );
      Continue( Item->ThreadData );
      break;
    case RELEASE_CB_RV2 :
      //TsmAssert( Item->ThreadData->State == WAITING_RV2 );
      Continue( Item->ThreadData );
      break;
    case RELEASE_MSG :
      //TsmAssert( Item->ThreadData->hasMessage() );
      Continue( Item->ThreadData );
      break;
    case RELEASE_RV2 :
      //TsmAssert( ( Item->ThreadData->State & WAITING_RV ) != 0 );
      Continue( Item->ThreadData );
      break;
    case NATIVE :
      ReplayNative( Item );
      break;
    case CHOICE :
      TsmRecorder::Enable();
      TsmRecorder::Show( "**** BUG " );
      throw TsmException( "TismBUG: Handled by 'ChoiceMax'" );
      //break;
    case QO :
      break;
    }
  }

  void TsmSvrChecker::Replay()
  {
    for( unsigned int i=0; i<Classes.size(); i++ )
    {
      Classes[i]->TsmReset2();
    }

    CurrentScenItemIndex = 0;
    TsmRecorder::Reset();

    for ( ; ; )
    {
      CheckAllThreadsWaiting();

      if ( CurrentScenItemIndex == (int) CurrentScenario->Items->size() )
      { /*
         * Ready replaying, possibly due to choice items, or
         * empty scenario.
         */
        TsmRecorder::Enable();
        break;
        /****/
      }
      if ( CurrentScenItemIndex == (int) CurrentScenario->Items->size()-1 )
      { /*
         * The last item is 'new'.
         */
        TsmRecorder::Enable();
      }
      boost::shared_ptr<TsmScenItem> SpSi = CurrentScenario->Items->at( CurrentScenItemIndex++ );
      ReplayScenItem( SpSi.get() );
    }
  }

  TsmScenItem* TsmSvrChecker::GetThreadDataRV2()
  {
    TsmScenItem* lScenItem;

    if ( WAITING_RV2 == CbThreadData->State )
    {
      lScenItem = new TsmScenItem();
      lScenItem->Kind = RELEASE_CB_RV2;
      lScenItem->ThreadData = CbThreadData;
      return lScenItem;
    }

    map<boost::thread::id,TsmThreadData*>::iterator it;
    for ( it=mThreadDatas.begin() ; it != mThreadDatas.end(); it++ )
    {
      TsmThreadData *lThreadData = it->second;

      if ( WAITING_RV2 == lThreadData->State )
      {
        lScenItem = new TsmScenItem();
        lScenItem->Kind = RELEASE_RV2;
        lScenItem->ThreadData = lThreadData;
        return lScenItem;
      }
    }
    
    return NULL;
  }

  void TsmSvrChecker::ReplayExtend( TsmScenario* oScenario )
  {
    TsmScenItem* lScenItem;

    while ( NULL != (lScenItem = GetThreadDataRV2()) )
    {
      oScenario->Add( lScenItem );
      CurrentScenItemIndex++;
      ReplayScenItem( lScenItem );
      CheckAllThreadsWaiting();
    }
  }

  void TsmSvrChecker::AddToCurrentScenario( vector<TsmScenItem*>& oScenItems )
  {
    for( unsigned int i=0; i<oScenItems.size(); i++ )
    {
      TsmScenario* lNewScenario = CurrentScenario->Extend( oScenItems[i] );
      FutureScenarios.push( lNewScenario );
    }
  }

  void TsmSvrChecker::AddScenarioNatives( vector<TsmScenItem*>& oScenItems, bool bClient )
  {
    boost::unique_lock<boost::mutex> lock( LockNatives );

    map<TsmClass*,TsmNativeLists*>::iterator it;

    for ( it = Natives.begin(); it != Natives.end(); it++ )
    {
      TsmNativeLists* lLists = it->second;
      TsmClass* Receiver = it->first;

      multiset<int>::iterator itm;
      for ( itm = lLists->ServerNatives.begin();
        itm != lLists->ServerNatives.end();
        itm++ )
      {
        TsmScenItem* lScenItem = new TsmScenItem();
        lScenItem->Kind = NATIVE;
        lScenItem->Index = *itm; // index;
        lScenItem->Receiver = Receiver;
        oScenItems.push_back( lScenItem );
      }

      if ( bClient )
      {
        for ( itm = lLists->ClientNatives.begin();
          itm != lLists->ClientNatives.end();
          itm++ )
        {
          TsmScenItem* lScenItem = new TsmScenItem();
          lScenItem->Kind = NATIVE;
          lScenItem->Index = *itm; // index;
          lScenItem->Receiver = Receiver;
          oScenItems.push_back( lScenItem );
        }
      }
    }
  }

  void TsmSvrChecker::AddScenarioCbThread( vector<TsmScenItem*>& oScenItems )
  {
    if ( CbThreadData->pTsmThread->getIntQueue()->size() == 0 )
    {
      return;
    }

    if ( CbThreadData->State == WAITING_RV )
    {
      return;
    }
    //TsmAssert( CbThreadData->State == WAITING_MSG );

    TsmScenItem* lScenItem = new TsmScenItem();
    lScenItem->Kind = RELEASE_CB_MSG;
    lScenItem->ThreadData = CbThreadData;
    oScenItems.push_back( lScenItem );
  }

  void TsmSvrChecker::AddScenarioThreads( vector<TsmScenItem*>& oScenItems )
  {
    map<boost::thread::id,TsmThreadData*>::iterator it;
    for ( it=mThreadDatas.begin() ; it != mThreadDatas.end(); it++ )
    {
      TsmThreadData *lThreadData = it->second;

      if ( WAITING_MSG == lThreadData->State )
      {
        if ( lThreadData->pTsmThread->hasMessage() )
        {
          TsmScenItem* lScenItem = new TsmScenItem();
          lScenItem->Kind = RELEASE_MSG;
          lScenItem->ThreadData = lThreadData;
          oScenItems.push_back( lScenItem );
        }
      }
    }
  }

  string TsmSvrChecker::GetDuration( long duration )
  {
    long s = duration % 60;
    duration /= 60;
    long m = duration % 60;
    duration /= 60;

    stringstream ss;
    ss.fill('0');
    ss.width(2);
    ss << duration << ":";
    ss.width(2);
    ss << m << ":";
    ss.width(2);
    ss << s;
    return ss.str();
  }

  string TsmSvrChecker::TxtInt( string s, int k )
  {
    stringstream ss;
    ss << s << k;
    return ss.str();
  }
}
