/*
 * TismTool © Tis Veugen, 2010-2012
 * Simulation, Verification
 *
 * TsmSvrChecker.h
 * version 3
 */
#ifndef TSM_SVRCHECKER_H
#define TSM_SVRCHECKER_H

#include <map>
#include <stack>
#include <vector>
#include <ctime>    // time
#include <boost/thread/thread.hpp>
#include "TsmTriggerSem.h"
#include "TsmThread.h"
#include "TsmThreadType.h"
#include "TsmThreadState.h"
#include "TsmClass.h"
#include "TsmSvrHelpers.h"
using namespace std;

namespace TsmRTE
{
  class TsmSvrChecker
  {
  private:
    TsmTriggerSem  mTriggerSem;
    TsmThreadData* MyThreadData;
    TsmThreadData* CbThreadData;
    map<boost::thread::id,TsmThreadData*> mThreadDatas;

    boost::mutex                   LockNatives;
    map<TsmClass*,TsmNativeLists*> Natives;

    map<string,TsmSnapdata*> Snapshots;
    vector<TsmClass*>     Classes;
    stack<TsmScenario*>   PastScenarios;
    stack<TsmScenario*>   FutureScenarios;
    TsmScenario*          CurrentScenario;
    int                   CurrentScenItemIndex;
    bool                  bQueueOverflow;

  private:
    int          NumberOfDeadlocks;
    int          NumberOfLivelocks;
    int          NumberOfCycles;
    int          NumberOfScenarios;
    long         NumberOfLoops;
    time_t       StartTime;
    time_t       StopTime;

  public:
    TsmSvrChecker()
      : CurrentScenItemIndex( 0 )
      , bQueueOverflow( false )
      , NumberOfDeadlocks( 0 )
      , NumberOfLivelocks( 0 )
      , NumberOfCycles( 0 )
      , NumberOfScenarios( 0 )
      , NumberOfLoops( 0 )
    {
    }

    static int            mThreadId;
    static TsmSvrChecker* SingleInstance;

    static TsmSvrChecker& Instance()
    {
      if ( ! SingleInstance )
      {
        SingleInstance = new TsmSvrChecker;
      }
      return *SingleInstance;
    }

    static void Cleanup()
    {
      if ( SingleInstance )
      {
        SingleInstance->DeleteThreadData();
        delete SingleInstance;
      }
    }

  public:
    vector<TsmClass*>* GetClasses()
    {
      return &Classes;
    }
    int Register( TsmThread* oTsmThread, TsmThreadType eType, boost::thread::id lThreadBid );
    void SetState( boost::thread::id ThreadBid, TsmThreadState eState );
    void NativeEntry( TsmClass* oOwner, const int aNatives[], int NrNatives );
    void NativeExit( TsmClass* oOwner, const int aNatives[], int NrNatives );
    void AddQueueOverflow();
    bool ChoiceOKVerify( int MaxNrChoices, int ChoiceIndex, string sGuard );
    void Verify();
    bool ChoiceOKSimulate( int MaxNrChoices, int ChoiceIndex );
    void Simulate();
    void Summary( bool bVerify );
    string GetCurrentSituation();
    string SnappedClasses();
    string SnappedThreads();
    string GetThreadStates( string sHeader );

  private:
    void DeleteScenItems( vector<TsmScenItem*>& oScenItems );
    int GetRandomInt( int iMax );
    bool AllThreadsWaiting();
    bool AllThreadsSleeping();
    void CheckAllThreadsWaiting();
    void Reset();
    void Free();
    void DeleteThreadData();
    void Pause( TsmThreadData* oThreadData );
    void Continue( TsmThreadData* oThreadData );
    void NativesReset();
    void ReplayNative( TsmScenItem* oItem );
    void ReplayScenItem( TsmScenItem* Item );
    void Replay();
    TsmScenItem* GetThreadDataRV2();
    void ReplayExtend( TsmScenario* oScenario );
    void AddToCurrentScenario( vector<TsmScenItem*>& oScenItems );
    void AddScenarioNatives( vector<TsmScenItem*>& oScenItems, bool bClient );
    void AddScenarioCbThread( vector<TsmScenItem*>& oScenItems );
    void AddScenarioThreads( vector<TsmScenItem*>& oScenItems );
    string GetDuration( long duration );
    string TxtInt( string s, int k );
  };
}
#endif
